
int read_int(FILE *in);
int read_word(FILE *in);

int read_chars(FILE *in, char *s, int count);

int read_int_b(FILE *in);
int read_word_b(FILE *in);

int read_buffer(FILE *in, unsigned char *buffer, int len);
